/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * Copyright (C) 2010 Red Hat, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Jan Görig <jgorig@redhat.com>
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dbus/dbus.h>
#include <sys/select.h>
#include <netinet/in.h>
#include "libnmserver-dbus.h"
#include "libnmserver-internal.h"

#define NM_PATH "/org/freedesktop/NetworkManager"
#define NM_BUS_NAME "org.freedesktop.NetworkManager"
#define NM_IF "org.freedesktop.NetworkManager"
#define NM_IF_DEVICE "org.freedesktop.NetworkManager.Device"
#define NM_IF_IP4CONFIG "org.freedesktop.NetworkManager.IP4Config"

fd_set rfds; /* File descriptors set for select */
int rfds_max; /* Max FD id in set */

/* Initialize D-Bus connection */
DBusConnection* nmserver_dbus_init(void)
{
	DBusError error;
	DBusConnection *conn;

	dbus_error_init(&error);

	conn = dbus_bus_get(DBUS_BUS_SYSTEM, &error);
	if (dbus_error_is_set(&error)) { /* XXX What about conn == NULL */
		dbus_error_free(&error);
		return NULL;
	}

	return conn;
}

/* Close D-Bus connection */
void nmserver_dbus_close(DBusConnection *conn)
{
	dbus_connection_unref(conn);
}

/* Returns D-Bus property as a message */
DBusMessage* nmserver_dbus_get_property(DBusConnection *conn,
					const char *target, const char *path,
					const char *interface, const char *name)
{
	DBusMessage *msg, *reply = NULL;
	DBusMessageIter args;
	DBusError error;

	dbus_error_init(&error);

	msg = dbus_message_new_method_call(target, path,
					   DBUS_INTERFACE_PROPERTIES, "Get");
	if (msg == NULL)
		return NULL;

	dbus_message_iter_init_append(msg, &args);

	if (!dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING,
					    &interface) ||
	    !dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING, &name))
		goto cleanup;

	reply = dbus_connection_send_with_reply_and_block(conn, msg, -1, &error);
	if (dbus_error_is_set(&error)) {
		dbus_error_free(&error);
		goto cleanup;
	}

cleanup:
	dbus_message_unref(msg);

	return reply;
}

/* Gets IPv4 nameservers addresses from Ip4Config NM object */
nmserver_result_t nmserver_dbus_get_nameservers(DBusConnection *conn,
						const char *path,
						in_addr_t **servers)
{
	DBusMessage *msg;
	DBusMessageIter iter;
	in_addr_t *addrs;
	int count;
	nmserver_result_t res = NMSERVER_R_BADMSG;

	msg = nmserver_dbus_get_property(conn, NM_BUS_NAME, path,
					 NM_IF_IP4CONFIG, "Nameservers");

	if (msg == NULL)
		return NMSERVER_R_NORESP;

	if (!dbus_message_iter_init(msg, &iter) ||
	    dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_VARIANT)
		goto cleanup;

	dbus_message_iter_recurse(&iter, &iter);

	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_ARRAY)
		goto cleanup;

	dbus_message_iter_recurse(&iter, &iter);

	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_UINT32)
		goto cleanup;

	dbus_message_iter_get_fixed_array(&iter, &addrs, &count);

	*servers = malloc((count + 1) * sizeof(in_addr_t));
	if (*servers == NULL) {
		res = NMSERVER_R_MEMORY;
		goto cleanup;
	}

	memcpy(*servers, addrs, count * sizeof(in_addr_t));
	(*servers)[count] = 0;

	res = NMSERVER_R_OK;

cleanup:
	dbus_message_unref(msg);

	return res;
}

/* Gets Ip4Config object path from Device NM object */
nmserver_result_t nmserver_dbus_get_ip4config(DBusConnection *conn,
					      const char *path,
					      char **ip4config)
{
	DBusMessage *msg;
	DBusMessageIter iter;
	const char *str;
	int len;
	nmserver_result_t res;

	msg = nmserver_dbus_get_property(conn, NM_BUS_NAME, path, NM_IF_DEVICE,
					 "Ip4Config");

	if (msg == NULL)
		return NMSERVER_R_NORESP;

	/* dbus_message_iter* failures are NMSERVER_R_BADMSG errors */
	res = NMSERVER_R_BADMSG;

	if (!dbus_message_iter_init(msg, &iter) ||
	    dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_VARIANT)
		goto cleanup;

	dbus_message_iter_recurse(&iter, &iter);

	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_OBJECT_PATH)
		goto cleanup;

	dbus_message_iter_get_basic(&iter, &str);

	len = strlen(str) + 1;

	*ip4config = malloc(len);
	if (*ip4config == NULL) {
		res = NMSERVER_R_MEMORY;
		goto cleanup;
	}

	memcpy(*ip4config, str, len);

	res = NMSERVER_R_OK;

cleanup:
	dbus_message_unref(msg);

	return res;
}

/* Frees structure returned by nmserver_dbus_get_devices */
void nmserver_dbus_free_devices(nmserver_devices_t *devices)
{
	for (int i = 0; i < devices->count; i++)
		free(devices->devices[i]);

	free(devices->devices);
}

/* Gets available network devices from NM */
nmserver_result_t nmserver_dbus_get_devices(DBusConnection *conn,
					    nmserver_devices_t *devices)
{
	DBusMessage *msg, *reply;
	DBusMessageIter iter, deviter;
	DBusError error;
	int count, len;
	char *str;
	nmserver_result_t res;

	dbus_error_init(&error);

	msg = dbus_message_new_method_call(NM_BUS_NAME, NM_PATH, NM_IF,
					   "GetDevices");
	if (msg == NULL)
		return NMSERVER_R_MEMORY;

	reply = dbus_connection_send_with_reply_and_block(conn, msg, -1, &error);

	dbus_message_unref(msg);

	if (dbus_error_is_set(&error)) { /* XXX What about reply == NULL ? */
		dbus_error_free(&error);
		return NMSERVER_R_NORESP;
	}

	if (reply == NULL)
		return NMSERVER_R_NORESP;

	/* dbus_message_iter* failures are NMSERVER_R_BADMSG errors */
	res = NMSERVER_R_BADMSG;

	if (!dbus_message_iter_init(reply, &iter) ||
	    dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_ARRAY)
		goto cleanup;

	dbus_message_iter_recurse(&iter, &deviter);

	if (dbus_message_iter_get_arg_type(&deviter) != DBUS_TYPE_OBJECT_PATH)
		goto cleanup;

	/* Count number of devices */
	count = 0;
	while (dbus_message_iter_get_arg_type(&deviter) != DBUS_TYPE_INVALID) {
		count++;
		dbus_message_iter_next(&deviter);
	}

	devices->devices = malloc(count * sizeof(char*));
	if (devices->devices == NULL) {
		res = NMSERVER_R_MEMORY;
		goto cleanup;
	}

	dbus_message_iter_recurse(&iter, &deviter);

	/* Save devices paths to structure */
	devices->count = 0;
	while (dbus_message_iter_get_arg_type(&deviter) != DBUS_TYPE_INVALID) {
		dbus_message_iter_get_basic(&deviter, &str);

		len = strlen(str) + 1;

		devices->devices[devices->count] = malloc(len);
		if (devices->devices[devices->count] == NULL) {
			nmserver_dbus_free_devices(devices);
			res = NMSERVER_R_MEMORY;
			goto cleanup;
		}

		memcpy(devices->devices[devices->count], str, len);
		devices->count++;

		dbus_message_iter_next(&deviter);
	}

	res = NMSERVER_R_OK;

cleanup:
	dbus_message_unref(reply);

	return res;
}

/* Gets nameservers addresses and calls callback */
nmserver_result_t nmserver_dbus_callback_nameservers(nmserver_control_t *control)
{
	char *ip4config;
	nslist_t nameservers;
	in_addr_t *addrs, *ptr;
	nmserver_result_t res;

	nslist_init(&nameservers);

	nmserver_devices_t devices;

	res = nmserver_dbus_get_devices(control->conn, &devices);
	if (res != NMSERVER_R_OK)
		return res;

	for (int i = 0; i < devices.count; i++) {
		res = nmserver_dbus_get_ip4config(control->conn,
						  devices.devices[i],
						  &ip4config);
		if (res != NMSERVER_R_OK) {
			nmserver_dbus_free_devices(&devices);
			return res;
		}

		/* Device doesn't have active connection */
		if (!strcmp(ip4config, "/"))
			continue;

		res = nmserver_dbus_get_nameservers(control->conn, ip4config,
						    &addrs);
		if (res != NMSERVER_R_OK) {
			free(ip4config);
			nmserver_dbus_free_devices(&devices);
			return res;
		}

		ptr = addrs;

		while (*ptr) {
			nslist_add(&nameservers, *ptr);
			ptr++;
		}

		free(addrs);
		free(ip4config);
	}

	nmserver_dbus_free_devices(&devices);

	/* If there is no change, free the list */
	if (nslist_compare(&nameservers, &control->nameservers))
		nslist_free(&nameservers);
	else {
		nslist_move(&control->nameservers, &nameservers);
		addrs = nslist_serialize(&control->nameservers);

		if (addrs == NULL)
			return NMSERVER_R_MEMORY;

		control->nscallback(addrs, control->data);

		free(addrs);
	}

	return NMSERVER_R_OK;
}

/* D-Bus watch add function */
/* XXX Why we return something here? */
dbus_bool_t nmserver_dbus_watch_add(DBusWatch *watch, void *data)
{
	int fd;

	/* Works only with read watches */
	if ((dbus_watch_get_flags(watch) & DBUS_WATCH_READABLE) != 0) {
		fd = dbus_watch_get_unix_fd(watch);
		FD_SET(fd, &rfds);

		if (rfds_max < fd)
			rfds_max = fd;
	}

	return TRUE;
}

/* D-Bus watch remove function */
void nmserver_dbus_watch_rem(DBusWatch *watch, void *data)
{
	/* TODO: Change rfds_max to lower value. */
	FD_CLR(dbus_watch_get_unix_fd(watch), &rfds);
}

/* D-Bus watch toggle function */
void nmserver_dbus_watch_toggle(DBusWatch *watch, void *data)
{
	if (dbus_watch_get_enabled(watch))
		nmserver_dbus_watch_add(watch, NULL);
	else
		nmserver_dbus_watch_rem(watch, NULL);
}

/* D-Bus main reading loop */
int nmserver_dbus_loop(nmserver_control_t *control)
{
	DBusError error;
	DBusMessage *msg;

	dbus_error_init(&error);

	dbus_bus_add_match(control->conn,
			   "type='signal',interface='" NM_IF_DEVICE "'",
			   &error);

	dbus_connection_flush(control->conn);

	if (dbus_error_is_set(&error))
		return NMSERVER_R_MATCHFAIL;

	FD_ZERO(&rfds);
	FD_SET(control->comm_pipe[0], &rfds);
	rfds_max = control->comm_pipe[0];

	if (!dbus_connection_set_watch_functions(control->conn,
						 nmserver_dbus_watch_add,
						 nmserver_dbus_watch_rem,
						 nmserver_dbus_watch_toggle,
						 NULL, free))
		return NMSERVER_R_MEMORY;

	fd_set crfds = rfds;

	while (select(rfds_max+1, &crfds, NULL, NULL, NULL)) {
		if (FD_ISSET(control->comm_pipe[0], &crfds))
			break;

		crfds = rfds;

		dbus_connection_read_write(control->conn, 0);
		msg = dbus_connection_pop_message(control->conn);

		if (msg == NULL) {
			sleep(1); /* XXX Why we sleep here? */
			continue;
		}

		/*
		 * Checking for StateChanged and NameAcquired signals.
		 * Sometimes StateChanged is not emited for first time,
		 * so check for NameAcquired.
		 */
		if (dbus_message_is_signal(msg, NM_IF_DEVICE, "StateChanged") ||
		    dbus_message_is_signal(msg, DBUS_INTERFACE_DBUS,
					   "NameAcquired")) {
			nmserver_result_t res;

			res = nmserver_dbus_callback_nameservers(control);

			if (res != NMSERVER_R_OK && control->errcallback != NULL)
				control->errcallback(res);
		}

		dbus_message_unref(msg);
	}

	close(control->comm_pipe[0]);

	return NMSERVER_R_OK;
}
